package cc.fascinated.bat.features.logging.listeners;

import cc.fascinated.bat.common.EmbedDescriptionBuilder;
import cc.fascinated.bat.common.EmbedUtils;
import cc.fascinated.bat.common.EnumUtils;
import cc.fascinated.bat.event.EventListener;
import cc.fascinated.bat.features.logging.LogFeature;
import cc.fascinated.bat.features.logging.LogType;
import cc.fascinated.bat.model.BatGuild;
import cc.fascinated.bat.model.BatUser;
import cc.fascinated.bat.service.DiscordService;
import cc.fascinated.bat.service.UserService;
import jakarta.annotation.PostConstruct;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import net.dv8tion.jda.api.entities.Member;
import net.dv8tion.jda.api.entities.channel.ChannelType;
import net.dv8tion.jda.api.entities.channel.concrete.TextChannel;
import net.dv8tion.jda.api.entities.channel.concrete.VoiceChannel;
import net.dv8tion.jda.api.entities.channel.unions.AudioChannelUnion;
import net.dv8tion.jda.api.entities.channel.unions.ChannelUnion;
import net.dv8tion.jda.api.events.channel.ChannelCreateEvent;
import net.dv8tion.jda.api.events.channel.ChannelDeleteEvent;
import net.dv8tion.jda.api.events.channel.update.ChannelUpdateNameEvent;
import net.dv8tion.jda.api.events.guild.voice.GenericGuildVoiceEvent;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Component;

import java.util.HashMap;
import java.util.Map;

/**
 * @author Fascinated (fascinated7)
 */
@Component
@Log4j2
public class ChannelListener implements EventListener {
    /**
     * A map of users and the last voice channel they were in
     */
    private final Map<BatUser, VoiceChannel> lastVoiceChannel = new HashMap<>();
    private final UserService userService;
    private final LogFeature logFeature;

    @Autowired
    public ChannelListener(@NonNull ApplicationContext context, @NonNull UserService userService) {
        this.userService = userService;
        this.logFeature = context.getBean(LogFeature.class);
    }

    @PostConstruct
    public void init() {
        log.info("Getting last voice channel for all users");
        for (VoiceChannel voiceChannel : DiscordService.JDA.getVoiceChannels()) {
            for (Member member : voiceChannel.getMembers()) {
                BatUser user = userService.getUser(member.getId());
                if (user == null) {
                    continue;
                }
                lastVoiceChannel.put(user, voiceChannel);
            }
        }
        log.info("Found {} users in a voice channel", lastVoiceChannel.size());
    }

    @Override
    public void onChannelCreate(@NonNull BatGuild guild, @NonNull ChannelCreateEvent event) {
        ChannelUnion channel = event.getChannel();
        String type = formatChannelType(channel);
        log.info("{} \"{}\" was created in guild \"{}\"", type, event.getChannel().getName(), guild.getName());
        logFeature.sendLog(guild, LogType.CHANNEL_CREATE, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("%s Created".formatted(type))
                        .appendLine("%s: %s".formatted(type, event.getChannel().getAsMention()), true)
                        .appendLine("Name: `%s`".formatted(event.getChannel().getName()), true)
                        .build())
                .build());
    }

    @Override
    public void onChannelDelete(@NonNull BatGuild guild, @NonNull ChannelDeleteEvent event) {
        ChannelUnion channel = event.getChannel();
        String type = formatChannelType(channel);
        log.info("{} \"{}\" was deleted in guild \"{}\"", type, event.getChannel().getName(), guild.getName());
        EmbedDescriptionBuilder description = new EmbedDescriptionBuilder("%s Deleted".formatted(type))
                .appendLine("Name: `#%s`".formatted(channel.getName()), true);
        if (channel instanceof TextChannel) {
            TextChannel textChannel = channel.asTextChannel();
            description.appendLine("Topic: `%s`".formatted(textChannel.getTopic() == null ? "No Topic" : textChannel.getTopic()), true);
        }
        logFeature.sendLog(guild, LogType.CHANNEL_DELETE, EmbedUtils.errorEmbed().setDescription(description.build()).build());
    }

    @Override
    public void onChannelUpdateName(@NonNull BatGuild guild, @NonNull ChannelUpdateNameEvent event) {
        ChannelUnion channel = event.getChannel();
        String type = formatChannelType(channel);
        log.info("{} \"{}\" name was updated to {} in guild \"{}\"", type, event.getNewValue(), event.getOldValue(), guild.getName());
        EmbedDescriptionBuilder description = new EmbedDescriptionBuilder("%s Name Updated".formatted(type));
        if (channel.getType() == ChannelType.TEXT || channel.getType() == ChannelType.VOICE) {
            description.appendLine("Channel: %s".formatted(event.getChannel().getAsMention()), true);
        }
        description.appendLine("Old Name: `%s`".formatted(event.getOldValue()), true);
        description.appendLine("New Name: `%s`".formatted(event.getNewValue()), true);
        logFeature.sendLog(guild, LogType.CHANNEL_UPDATE_NAME, EmbedUtils.successEmbed()
                .setDescription(description
                        .build())
                .build());
    }

    @Override
    public void onGuildVoiceUpdate(@NonNull BatGuild guild, @NonNull BatUser user, @NonNull GenericGuildVoiceEvent event) {
        AudioChannelUnion channel = event.getVoiceState().getChannel();
        if (channel != null) {
            VoiceChannel voiceChannel = channel.asVoiceChannel();
            lastVoiceChannel.put(user, voiceChannel);
        }
        VoiceChannel voiceChannel = lastVoiceChannel.get(user);
        if (voiceChannel == null) {
            return;
        }
        boolean joined = voiceChannel.getMembers().contains(event.getMember());
        if (!joined) {
            lastVoiceChannel.remove(user);
        }
        log.info("User \"{}\" {} voice channel \"{}\" in guild \"{}\"", user.getName(), joined ? "joined" : "left", voiceChannel.getName(), guild.getName());
        String description = new EmbedDescriptionBuilder("Member %s Voice Channel".formatted(joined ? "Joined" : "Left"))
                .appendLine("User: %s".formatted(user.getDiscordUser().getAsMention()), true)
                .appendLine("Channel: %s".formatted(voiceChannel.getAsMention()), true)
                .build();
        if (joined) {
            logFeature.sendLog(guild, LogType.VOICE_CHANNEL_JOIN, EmbedUtils.successEmbed().setDescription(description).build());
            return;
        }
        logFeature.sendLog(guild, LogType.VOICE_CHANNEL_LEAVE, EmbedUtils.errorEmbed().setDescription(description).build());
    }

    /**
     * Formats the channel type
     *
     * @param channel - the channel
     * @return the formatted channel type
     */
    public String formatChannelType(ChannelUnion channel) {
        return (channel.getType() == ChannelType.TEXT || channel.getType() == ChannelType.VOICE) ? EnumUtils.getEnumName(channel.getType()) + " Channel"
                : EnumUtils.getEnumName(channel.getType());
    }
}
