package cc.fascinated.bat.features.logging.listeners;

import cc.fascinated.bat.common.EmbedDescriptionBuilder;
import cc.fascinated.bat.common.EmbedUtils;
import cc.fascinated.bat.event.EventListener;
import cc.fascinated.bat.features.logging.LogFeature;
import cc.fascinated.bat.features.logging.LogType;
import cc.fascinated.bat.model.BatGuild;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import net.dv8tion.jda.api.entities.Invite;
import net.dv8tion.jda.api.entities.emoji.Emoji;
import net.dv8tion.jda.api.events.emoji.EmojiAddedEvent;
import net.dv8tion.jda.api.events.emoji.EmojiRemovedEvent;
import net.dv8tion.jda.api.events.emoji.update.EmojiUpdateNameEvent;
import net.dv8tion.jda.api.events.guild.invite.GuildInviteCreateEvent;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Component;

/**
 * @author Fascinated (fascinated7)
 */
@Component
@Log4j2
public class GuildListener implements EventListener {
    private final LogFeature logFeature;

    @Autowired
    public GuildListener(@NonNull ApplicationContext context) {
        this.logFeature = context.getBean(LogFeature.class);
    }

    @Override
    public void onGuildInviteCreate(@NonNull BatGuild guild, @NonNull Invite invite, @NonNull GuildInviteCreateEvent event) {
        log.info("Invite created in guild \"{}\" with code \"{}\"", guild.getGuild().getName(), invite.getCode());

        guild.getDiscordGuild().retrieveInvites().queue(invites -> {
            Invite newInvite = invites.stream().filter(i -> i.getCode().equals(invite.getCode())).findFirst().orElse(null);
            if (newInvite == null) {
                return;
            }
            newInvite.expand().queue(expanded -> logFeature.sendLog(guild, LogType.INVITE_CREATE, EmbedUtils.successEmbed()
                    .setDescription(new EmbedDescriptionBuilder("Invite Created")
                            .appendLine("Invite: %s".formatted(expanded.getUrl()), true)
                            .appendLine("Channel: %s".formatted(expanded.getChannel()), true)
                            .appendLine("Max Uses: `%s`".formatted(expanded.getMaxUses() == 0 ? "Infinite" : expanded.getMaxUses()), true)
                            .appendLine("Creator: %s".formatted(expanded.getInviter() == null ? "Unknown" : expanded.getInviter().getAsMention()), true)
                            .build())
                    .build()));
        });
    }

    @Override
    public void onEmojiAdd(@NonNull BatGuild guild, @NonNull Emoji emoji, @NonNull EmojiAddedEvent event) {
        log.info("Emoji \"{}\" was added in guild \"{}\"", emoji.getName(), guild.getName());
        logFeature.sendLog(guild, LogType.EMOJI_ADD, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("Emoji Added")
                        .appendLine("Emoji: %s".formatted(emoji.getFormatted()), true)
                        .appendLine("Name: `%s`".formatted(emoji.getName()), true)
                        .build())
                .build());
    }

    @Override
    public void onEmojiRemove(@NonNull BatGuild guild, @NonNull Emoji emoji, @NonNull EmojiRemovedEvent event) {
        log.info("Emoji \"{}\" was removed in guild \"{}\"", emoji.getName(), guild.getName());
        logFeature.sendLog(guild, LogType.EMOJI_REMOVE, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("Emoji Removed")
                        .appendLine("Emoji: %s".formatted(emoji.getFormatted()), true)
                        .appendLine("Name: `%s`".formatted(emoji.getName()), true)
                        .build())
                .build());
    }

    @Override
    public void onEmojiRename(@NonNull BatGuild guild, @NonNull Emoji emoji, @NonNull String oldName, @NonNull String newName, @NonNull EmojiUpdateNameEvent event) {
        log.info("Emoji \"{}\" was renamed to \"{}\" in guild \"{}\"", oldName, newName, guild.getName());
        logFeature.sendLog(guild, LogType.EMOJI_RENAME, EmbedUtils.successEmbed()
                .setDescription(new EmbedDescriptionBuilder("Emoji Renamed")
                        .appendLine("Emoji: %s".formatted(emoji.getFormatted()), true)
                        .appendLine("Old Name: `%s`".formatted(oldName), true)
                        .appendLine("New Name: `%s`".formatted(newName), true)
                        .build())
                .build());
    }
}
